@@ -1,27 +1,46 @@
 """Support for various GEOS geometry operations
 """
 
-from ctypes import byref, c_void_p, c_double
 from warnings import warn
 
+import shapely
+from shapely.algorithms.polylabel import polylabel  # noqa
 from shapely.errors import GeometryTypeError, ShapelyDeprecationWarning
-from shapely.prepared import prep
-from shapely.geos import lgeos
-from shapely.geometry.base import geom_factory, BaseGeometry, BaseMultipartGeometry
 from shapely.geometry import (
-    shape, Point, MultiPoint, LineString, MultiLineString, Polygon, GeometryCollection)
+    GeometryCollection,
+    LineString,
+    MultiLineString,
+    MultiPoint,
+    Point,
+    Polygon,
+    shape,
+)
+from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry
 from shapely.geometry.polygon import orient as orient_
-from shapely.algorithms.polylabel import polylabel
+from shapely.prepared import prep
 
-
-__all__ = ['cascaded_union', 'linemerge', 'operator', 'polygonize',
-           'polygonize_full', 'transform', 'unary_union', 'triangulate',
-           'voronoi_diagram', 'split', 'nearest_points', 'validate', 'snap',
-           'shared_paths', 'clip_by_rect', 'orient', 'substring']
+__all__ = [
+    "cascaded_union",
+    "linemerge",
+    "operator",
+    "polygonize",
+    "polygonize_full",
+    "transform",
+    "unary_union",
+    "triangulate",
+    "voronoi_diagram",
+    "split",
+    "nearest_points",
+    "validate",
+    "snap",
+    "shared_paths",
+    "clip_by_rect",
+    "orient",
+    "substring",
+]
 
 
 class CollectionOperator:
-
     def shapeup(self, ob):
         if isinstance(ob, BaseGeometry):
             return ob
@@ -37,24 +56,15 @@ class CollectionOperator:
         The source may be a MultiLineString, a sequence of LineString objects,
         or a sequence of objects than can be adapted to LineStrings.
         """
-        source = getattr(lines, 'geoms', None) or lines
+        source = getattr(lines, "geoms", None) or lines
         try:
             source = iter(source)
         except TypeError:
             source = [source]
         finally:
-            obs = [self.shapeup(l) for l in source]
-        geom_array_type = c_void_p * len(obs)
-        geom_array = geom_array_type()
-        for i, line in enumerate(obs):
-            geom_array[i] = line._geom
-        product = lgeos.GEOSPolygonize(byref(geom_array), len(obs))
-        collection = geom_factory(product)
-        for g in collection.geoms:
-            clone = lgeos.GEOSGeom_clone(g._geom)
-            g = geom_factory(clone)
-            g._other_owned = False
-            yield g
+            obs = [self.shapeup(line) for line in source]
+        collection = shapely.polygonize(obs)
+        return collection.geoms
 
     def polygonize_full(self, lines):
         """Creates polygons from a source of lines, returning the polygons
@@ -63,7 +73,7 @@ class CollectionOperator:
         The source may be a MultiLineString, a sequence of LineString objects,
         or a sequence of objects than can be adapted to LineStrings.
 
-        Returns a tuple of objects: (polygons, dangles, cut edges, invalid ring
+        Returns a tuple of objects: (polygons, cut edges, dangles, invalid ring
         lines). Each are a geometry collection.
 
         Dangles are edges which have one or both ends which are not incident on
@@ -71,31 +81,16 @@ class CollectionOperator:
         form part of polygon. Invalid ring lines form rings which are invalid
         (bowties, etc).
         """
-        source = getattr(lines, 'geoms', None) or lines
+        source = getattr(lines, "geoms", None) or lines
         try:
             source = iter(source)
         except TypeError:
             source = [source]
         finally:
-            obs = [self.shapeup(l) for l in source]
-        L = len(obs)
-        subs = (c_void_p * L)()
-        for i, g in enumerate(obs):
-            subs[i] = g._geom
-        collection = lgeos.GEOSGeom_createCollection(5, subs, L)
-        dangles = c_void_p()
-        cuts = c_void_p()
-        invalids = c_void_p()
-        product = lgeos.GEOSPolygonize_full(
-            collection, byref(dangles), byref(cuts), byref(invalids))
-        return (
-            geom_factory(product),
-            geom_factory(dangles),
-            geom_factory(cuts),
-            geom_factory(invalids)
-            )
+            obs = [self.shapeup(line) for line in source]
+        return shapely.polygonize_full(obs)
 
-    def linemerge(self, lines):
+    def linemerge(self, lines, directed=False):
         """Merges all connected lines from a source
 
         The source may be a MultiLineString, a sequence of LineString objects,
@@ -103,62 +98,42 @@ class CollectionOperator:
         LineString or MultiLineString when lines are not contiguous.
         """
         source = None
-        if hasattr(lines, 'type') and lines.type == 'MultiLineString':
+        if getattr(lines, "geom_type", None) == "MultiLineString":
             source = lines
-        elif hasattr(lines, 'geoms'):
+        elif hasattr(lines, "geoms"):
             # other Multi geometries
             source = MultiLineString([ls.coords for ls in lines.geoms])
-        elif hasattr(lines, '__iter__'):
+        elif hasattr(lines, "__iter__"):
             try:
                 source = MultiLineString([ls.coords for ls in lines])
             except AttributeError:
                 source = MultiLineString(lines)
         if source is None:
-            raise ValueError("Cannot linemerge %s" % lines)
-        result = lgeos.GEOSLineMerge(source._geom)
-        return geom_factory(result)
+            raise ValueError(f"Cannot linemerge {lines}")
+        return shapely.line_merge(source, directed=directed)
 
     def cascaded_union(self, geoms):
         """Returns the union of a sequence of geometries
 
-        This function is deprecated, as it was superseded by
-        :meth:`unary_union`.
+        .. deprecated:: 1.8
+            This function was superseded by :meth:`unary_union`.
         """
         warn(
             "The 'cascaded_union()' function is deprecated. "
             "Use 'unary_union()' instead.",
-            ShapelyDeprecationWarning, stacklevel=2)
-        try:
-            if isinstance(geoms, BaseMultipartGeometry):
-                geoms = geoms.geoms
-            L = len(geoms)
-        except TypeError:
-            geoms = [geoms]
-            L = 1
-        subs = (c_void_p * L)()
-        for i, g in enumerate(geoms):
-            subs[i] = g._geom
-        collection = lgeos.GEOSGeom_createCollection(6, subs, L)
-        return geom_factory(lgeos.methods['cascaded_union'](collection))
+            ShapelyDeprecationWarning,
+            stacklevel=2,
+        )
+        return shapely.union_all(geoms, axis=None)
 
     def unary_union(self, geoms):
         """Returns the union of a sequence of geometries
 
-        This method replaces :meth:`cascaded_union` as the
-        preferred method for dissolving many polygons.
+        Usually used to convert a collection into the smallest set of polygons
+        that cover the same area.
         """
-        try:
-            if isinstance(geoms, BaseMultipartGeometry):
-                geoms = geoms.geoms
-            L = len(geoms)
-        except TypeError:
-            geoms = [geoms]
-            L = 1
-        subs = (c_void_p * L)()
-        for i, g in enumerate(geoms):
-            subs[i] = g._geom
-        collection = lgeos.GEOSGeom_createCollection(6, subs, L)
-        return geom_factory(lgeos.methods['unary_union'](collection))
+        return shapely.union_all(geoms, axis=None)
+
 
 operator = CollectionOperator()
 polygonize = operator.polygonize
@@ -183,9 +158,8 @@ def triangulate(geom, tolerance=0.0, edges=False):
     Otherwise the list of LineString edges is returned.
 
     """
-    func = lgeos.methods['delaunay_triangulation']
-    gc = geom_factory(func(geom._geom, tolerance, int(edges)))
-    return [g for g in gc.geoms]
+    collection = shapely.delaunay_triangles(geom, tolerance=tolerance, only_edges=edges)
+    return [g for g in collection.geoms]
 
 
 def voronoi_diagram(geom, envelope=None, tolerance=0.0, edges=False):
@@ -230,26 +204,24 @@ def voronoi_diagram(geom, envelope=None, tolerance=0.0, edges=False):
     [1] https://en.wikipedia.org/wiki/Voronoi_diagram
     [2] https://geos.osgeo.org/doxygen/geos__c_8h_source.html  (line 730)
     """
-    func = lgeos.methods['voronoi_diagram']
-    envelope = envelope._geom if envelope else None
     try:
-        result = geom_factory(func(geom._geom, envelope, tolerance, int(edges)))
-    except ValueError:
-        errstr = "Could not create Voronoi Diagram with the specified inputs."
+        result = shapely.voronoi_polygons(
+            geom, tolerance=tolerance, extend_to=envelope, only_edges=edges
+        )
+    except shapely.GEOSException as err:
+        errstr = "Could not create Voronoi Diagram with the specified inputs "
+        errstr += f"({err!s})."
         if tolerance:
             errstr += " Try running again with default tolerance value."
-        raise ValueError(errstr)
+        raise ValueError(errstr) from err
 
-    if result.type != 'GeometryCollection':
+    if result.geom_type != "GeometryCollection":
         return GeometryCollection([result])
     return result
 
 
-class ValidateOp:
-    def __call__(self, this):
-        return lgeos.GEOSisValidReason(this._geom)
-
-validate = ValidateOp()
+def validate(geom):
+    return shapely.is_valid_reason(geom)
 
 
 def transform(func, geom):
@@ -291,38 +263,40 @@ def transform(func, geom):
     """
     if geom.is_empty:
         return geom
-    if geom.type in ('Point', 'LineString', 'LinearRing', 'Polygon'):
+    if geom.geom_type in ("Point", "LineString", "LinearRing", "Polygon"):
 
         # First we try to apply func to x, y, z sequences. When func is
         # optimized for sequences, this is the fastest, though zipping
         # the results up to go back into the geometry constructors adds
         # extra cost.
         try:
-            if geom.type in ('Point', 'LineString', 'LinearRing'):
+            if geom.geom_type in ("Point", "LineString", "LinearRing"):
                 return type(geom)(zip(*func(*zip(*geom.coords))))
-            elif geom.type == 'Polygon':
-                shell = type(geom.exterior)(
-                    zip(*func(*zip(*geom.exterior.coords))))
-                holes = list(type(ring)(zip(*func(*zip(*ring.coords))))
-                             for ring in geom.interiors)
+            elif geom.geom_type == "Polygon":
+                shell = type(geom.exterior)(zip(*func(*zip(*geom.exterior.coords))))
+                holes = list(
+                    type(ring)(zip(*func(*zip(*ring.coords))))
+                    for ring in geom.interiors
+                )
                 return type(geom)(shell, holes)
 
         # A func that assumes x, y, z are single values will likely raise a
         # TypeError, in which case we'll try again.
         except TypeError:
-            if geom.type in ('Point', 'LineString', 'LinearRing'):
+            if geom.geom_type in ("Point", "LineString", "LinearRing"):
                 return type(geom)([func(*c) for c in geom.coords])
-            elif geom.type == 'Polygon':
-                shell = type(geom.exterior)(
-                    [func(*c) for c in geom.exterior.coords])
-                holes = list(type(ring)([func(*c) for c in ring.coords])
-                             for ring in geom.interiors)
+            elif geom.geom_type == "Polygon":
+                shell = type(geom.exterior)([func(*c) for c in geom.exterior.coords])
+                holes = list(
+                    type(ring)([func(*c) for c in ring.coords])
+                    for ring in geom.interiors
+                )
                 return type(geom)(shell, holes)
 
-    elif geom.type.startswith('Multi') or geom.type == 'GeometryCollection':
+    elif geom.geom_type.startswith("Multi") or geom.geom_type == "GeometryCollection":
         return type(geom)([transform(func, part) for part in geom.geoms])
     else:
-        raise GeometryTypeError('Type %r not recognized' % geom.type)
+        raise GeometryTypeError(f"Type {geom.geom_type!r} not recognized")
 
 
 def nearest_points(g1, g2):
@@ -330,35 +304,21 @@ def nearest_points(g1, g2):
 
     The points are returned in the same order as the input geometries.
     """
-    seq = lgeos.methods['nearest_points'](g1._geom, g2._geom)
+    seq = shapely.shortest_line(g1, g2)
     if seq is None:
         if g1.is_empty:
-            raise ValueError('The first input geometry is empty')
+            raise ValueError("The first input geometry is empty")
         else:
-            raise ValueError('The second input geometry is empty')
+            raise ValueError("The second input geometry is empty")
 
-    try:
-        x1 = c_double()
-        y1 = c_double()
-        x2 = c_double()
-        y2 = c_double()
-        lgeos.GEOSCoordSeq_getX(seq, 0, byref(x1))
-        lgeos.GEOSCoordSeq_getY(seq, 0, byref(y1))
-        lgeos.GEOSCoordSeq_getX(seq, 1, byref(x2))
-        lgeos.GEOSCoordSeq_getY(seq, 1, byref(y2))
-    finally:
-        lgeos._lgeos.GEOSCoordSeq_destroy(seq)
-
-    p1 = Point(x1.value, y1.value)
-    p2 = Point(x2.value, y2.value)
+    p1 = shapely.get_point(seq, 0)
+    p2 = shapely.get_point(seq, 1)
     return (p1, p2)
 
-def snap(g1, g2, tolerance):
-    """Snap one geometry to another with a given tolerance
 
-    Vertices of the first geometry are snapped to vertices of the second
-    geometry. The resulting snapped geometry is returned. The input geometries
-    are not modified.
+def snap(g1, g2, tolerance):
+    """
+    Snaps an input geometry (g1) to reference (g2) geometry's vertices.
 
     Parameters
     ----------
@@ -369,15 +329,11 @@ def snap(g1, g2, tolerance):
     tolerance : float
         The snapping tolerance
 
-    Example
-    -------
-    >>> square = Polygon([(1,1), (2, 1), (2, 2), (1, 2), (1, 1)])
-    >>> line = LineString([(0,0), (0.8, 0.8), (1.8, 0.95), (2.6, 0.5)])
-    >>> result = snap(line, square, 0.5)
-    >>> result.wkt
-    'LINESTRING (0 0, 1 1, 2 1, 2.6 0.5)'
+    Refer to :func:`shapely.snap` for full documentation.
     """
-    return(geom_factory(lgeos.methods['snap'](g1._geom, g2._geom, tolerance)))
+
+    return shapely.snap(g1, g2, tolerance)
+
 
 def shared_paths(g1, g2):
     """Find paths shared between the two given lineal geometries
@@ -399,11 +355,10 @@ def shared_paths(g1, g2):
         raise GeometryTypeError("First geometry must be a LineString")
     if not isinstance(g2, LineString):
         raise GeometryTypeError("Second geometry must be a LineString")
-    return(geom_factory(lgeos.methods['shared_paths'](g1._geom, g2._geom)))
+    return shapely.shared_paths(g1, g2)
 
 
 class SplitOp:
-
     @staticmethod
     def _split_polygon_with_line(poly, splitter):
         """Split a Polygon with a LineString"""
@@ -422,20 +377,26 @@ class SplitOp:
         # some polygonized geometries may be holes, we do not want them
         # that's why we test if the original polygon (poly) contains
         # an inner point of polygonized geometry (pg)
-        return [pg for pg in polygonize(union) if poly.contains(pg.representative_point())]
+        return [
+            pg for pg in polygonize(union) if poly.contains(pg.representative_point())
+        ]
 
     @staticmethod
     def _split_line_with_line(line, splitter):
         """Split a LineString with another (Multi)LineString or (Multi)Polygon"""
 
         # if splitter is a polygon, pick it's boundary
-        if splitter.type in ('Polygon', 'MultiPolygon'):
+        if splitter.geom_type in ("Polygon", "MultiPolygon"):
             splitter = splitter.boundary
 
         if not isinstance(line, LineString):
             raise GeometryTypeError("First argument must be a LineString")
-        if not isinstance(splitter, LineString) and not isinstance(splitter, MultiLineString):
-            raise GeometryTypeError("Second argument must be either a LineString or a MultiLineString")
+        if not isinstance(splitter, LineString) and not isinstance(
+            splitter, MultiLineString
+        ):
+            raise GeometryTypeError(
+                "Second argument must be either a LineString or a MultiLineString"
+            )
 
         # |    s\l   | Interior | Boundary | Exterior |
         # |----------|----------|----------|----------|
@@ -443,10 +404,10 @@ class SplitOp:
         # | Boundary |  0 or F  |    *     |    *     |   So either '0********' or '[0F]**0*****'
         # | Exterior |    *     |    *     |    *     |   No overlapping interiors ('1********')
         relation = splitter.relate(line)
-        if relation[0] == '1':
+        if relation[0] == "1":
             # The lines overlap at some segment (linear intersection of interiors)
-            raise ValueError('Input geometry segment overlaps with the splitter.')
-        elif relation[0] == '0' or relation[3] == '0':
+            raise ValueError("Input geometry segment overlaps with the splitter.")
+        elif relation[0] == "0" or relation[3] == "0":
             # The splitter crosses or touches the line's interior --> return multilinestring from the split
             return line.difference(splitter)
         else:
@@ -462,7 +423,7 @@ class SplitOp:
             raise GeometryTypeError("Second argument must be a Point")
 
         # check if point is in the interior of the line
-        if not line.relate_pattern(splitter, '0********'):
+        if not line.relate_pattern(splitter, "0********"):
             # point not on line interior --> return collection with single identity line
             # (REASONING: Returning a list with the input line reference and creating a
             # GeometryCollection at the general split function prevents unnecessary copying
@@ -477,28 +438,24 @@ class SplitOp:
         coords = list(line.coords)
         # split the line at the point and create two new lines
         current_position = 0.0
-        for i in range(len(coords)-1):
+        for i in range(len(coords) - 1):
             point1 = coords[i]
-            point2 = coords[i+1]
+            point2 = coords[i + 1]
             dx = point1[0] - point2[0]
             dy = point1[1] - point2[1]
-            segment_length = (dx ** 2 + dy ** 2) ** 0.5
+            segment_length = (dx**2 + dy**2) ** 0.5
             current_position += segment_length
             if distance_on_line == current_position:
                 # splitter is exactly on a vertex
-                return [
-                    LineString(coords[:i+2]),
-                    LineString(coords[i+1:])
-                ]
+                return [LineString(coords[: i + 2]), LineString(coords[i + 1 :])]
             elif distance_on_line < current_position:
                 # splitter is between two vertices
                 return [
-                    LineString(coords[:i+1] + [splitter.coords[0]]),
-                    LineString([splitter.coords[0]] + coords[i+1:])
+                    LineString(coords[: i + 1] + [splitter.coords[0]]),
+                    LineString([splitter.coords[0]] + coords[i + 1 :]),
                 ]
         return [line]
 
-
     @staticmethod
     def _split_line_with_multipoint(line, splitter):
         """Split a LineString with a MultiPoint"""
@@ -549,30 +506,44 @@ class SplitOp:
         'GEOMETRYCOLLECTION (LINESTRING (0 0, 1 1), LINESTRING (1 1, 2 2))'
         """
 
-        if geom.type in ('MultiLineString', 'MultiPolygon'):
-             return GeometryCollection([i for part in geom.geoms for i in SplitOp.split(part, splitter).geoms])
+        if geom.geom_type in ("MultiLineString", "MultiPolygon"):
+            return GeometryCollection(
+                [i for part in geom.geoms for i in SplitOp.split(part, splitter).geoms]
+            )
 
-        elif geom.type == 'LineString':
-            if splitter.type in ('LineString', 'MultiLineString', 'Polygon', 'MultiPolygon'):
+        elif geom.geom_type == "LineString":
+            if splitter.geom_type in (
+                "LineString",
+                "MultiLineString",
+                "Polygon",
+                "MultiPolygon",
+            ):
                 split_func = SplitOp._split_line_with_line
-            elif splitter.type in ('Point'):
+            elif splitter.geom_type == "Point":
                 split_func = SplitOp._split_line_with_point
-            elif splitter.type in ('MultiPoint'):
-                split_func =  SplitOp._split_line_with_multipoint
+            elif splitter.geom_type == "MultiPoint":
+                split_func = SplitOp._split_line_with_multipoint
             else:
-                raise GeometryTypeError("Splitting a LineString with a %s is not supported" % splitter.type)
+                raise GeometryTypeError(
+                    f"Splitting a LineString with a {splitter.geom_type} is not supported"
+                )
 
-        elif geom.type == 'Polygon':
-            if splitter.type == 'LineString':
+        elif geom.geom_type == "Polygon":
+            if splitter.geom_type == "LineString":
                 split_func = SplitOp._split_polygon_with_line
             else:
-                raise GeometryTypeError("Splitting a Polygon with a %s is not supported" % splitter.type)
+                raise GeometryTypeError(
+                    f"Splitting a Polygon with a {splitter.geom_type} is not supported"
+                )
 
         else:
-            raise GeometryTypeError("Splitting %s geometry is not supported" % geom.type)
+            raise GeometryTypeError(
+                f"Splitting {geom.geom_type} geometry is not supported"
+            )
 
         return GeometryCollection(split_func(geom, splitter))
 
+
 split = SplitOp.split
 
 
@@ -636,7 +607,10 @@ def substring(geom, start_dist, end_dist, normalized=False):
     """
 
     if not isinstance(geom, LineString):
-        raise GeometryTypeError("Can only calculate a substring of LineString geometries. A %s was provided." % geom.type)
+        raise GeometryTypeError(
+            "Can only calculate a substring of LineString geometries. "
+            f"A {geom.geom_type} was provided."
+        )
 
     # Filter out cases in which to return a point
     if start_dist == end_dist:
@@ -665,8 +639,8 @@ def substring(geom, start_dist, end_dist, normalized=False):
 
     if start_dist < 0:
         start_dist = geom.length + start_dist  # Values may still be negative,
-    if end_dist < 0:                           # but only in the out-of-range
-        end_dist = geom.length + end_dist      # sense, not the wrap-around sense.
+    if end_dist < 0:  # but only in the out-of-range
+        end_dist = geom.length + end_dist  # sense, not the wrap-around sense.
 
     reverse = start_dist > end_dist
     if reverse:
@@ -676,9 +650,9 @@ def substring(geom, start_dist, end_dist, normalized=False):
         start_dist = 0  # to avoid duplicating the first vertex
 
     if reverse:
-        vertex_list = [(end_point.x, end_point.y)]
+        vertex_list = [tuple(*end_point.coords)]
     else:
-        vertex_list = [(start_point.x, start_point.y)]
+        vertex_list = [tuple(*start_point.coords)]
 
     coords = list(geom.coords)
     current_distance = 0
@@ -691,11 +665,11 @@ def substring(geom, start_dist, end_dist, normalized=False):
         current_distance += ((p2[0] - p1[0]) ** 2 + (p2[1] - p1[1]) ** 2) ** 0.5
 
     if reverse:
-        vertex_list.append((start_point.x, start_point.y))
+        vertex_list.append(tuple(*start_point.coords))
         # reverse direction result
         vertex_list = reversed(vertex_list)
     else:
-        vertex_list.append((end_point.x, end_point.y))
+        vertex_list.append(tuple(*end_point.coords))
 
     return LineString(vertex_list)
 
@@ -727,8 +701,7 @@ def clip_by_rect(geom, xmin, ymin, xmax, ymax):
     """
     if geom.is_empty:
         return geom
-    result = geom_factory(lgeos.methods['clip_by_rect'](geom._geom, xmin, ymin, xmax, ymax))
-    return result
+    return shapely.clip_by_rect(geom, xmin, ymin, xmax, ymax)
 
 
 def orient(geom, sign=1.0):
